@@ -23,6 +23,7 @@ require 'json'                  # for math
 require 'rexml/document'        # for SVG and bibxml acrobatics
 
 require 'kramdown-rfc/doi'      # for fetching information for a DOI
+require 'kramdown-rfc/rfc8792'
 
 class Object
   def deep_clone
@@ -32,6 +33,53 @@ end
 
 module Kramdown
 
+  Kramdown::Options.define(:ol_start_at_first_marker, Kramdown::Options::Boolean, false, <<~EOF)
+      If this option is `true`, an ordered list (<ol) will use the
+      number in its first marker (1 for 1. etc.) as the default value
+      of the start= attribute.
+
+      Default: false (for backward compatibility)
+      Used by: RFCXML converter
+    EOF
+
+  Kramdown::Options.define(:nested_ol_types, Object, %w[1], <<~EOF) do |val|
+      Values for type= attribute for nested ordered lists (ol).
+      The value needs to be an array of <ol type= values, expressed as one of:
+      1. A YAML array
+      2. A string that will be split on commas (with optional blank space following)
+      3. A string that will be split on blank space
+
+      Default: ["1"]
+      Used by: RFCXML converter
+    EOF
+    val = case val
+          when String
+            if val[0] == "[" && val[-1] == "]"
+              begin
+                val = YAML.safe_load(val)
+              rescue Psych::SyntaxError
+                warn "** YAML syntax error in nested_ol_types=#{val.inspect}"
+                val = %w[1]
+              end
+            else
+              val = val.split(/, */)
+              val = val[0].split(/ +/) if val.size == 1
+            end
+            Kramdown::Options.simple_array_validator(val, :nested_ol_types)
+          when Array
+            val.map!{ |x| x.to_s }
+            val = Kramdown::Options.simple_array_validator(val, :nested_ol_types)
+          else
+            raise Kramdown::Error, "Invalid value for option '#{:nested_ol_types}': '#{val.inspect}'"
+          end
+    if val == []
+      val = %w[1]
+      warn "** Option #{:nested_ol_types} cannot be empty, defaulting to #{val.inspect}"
+    end
+    val
+  end
+
+
   module Parser
 
     class RFC2629Kramdown < Kramdown
@@ -41,8 +89,8 @@ module Kramdown
           sorted_abbrevs = @root.options[:abbrev_defs].keys.sort {|a, b| b.length <=> a.length }
           regexps = [Regexp.union(*sorted_abbrevs.map {|k|
                                     /#{Regexp.escape(k).gsub(/\\\s/, "[\\s\\p{Z}]+").force_encoding(Encoding::UTF_8)}/})]
-          # warn regexps.inspect
           regexps << /(?=(?:\W|^)#{regexps.first}(?!\w))/ # regexp should only match on word boundaries
+          # warn regexps.inspect
         end
         super(el, regexps)
       end
@@ -77,6 +125,11 @@ module Kramdown
         href.gsub(/\A(?:[0-9]|section-|u-|figure-|table-|iref-)/) { "_#{$&}" }
       end
 
+      def rfc_mention(target1)  # only works for RFCnnnn
+        target1 =~ /\A([A-Z]*)(.*)\z/
+        "#$1 #$2 "
+      end
+
       def handle_bares(s, attr, format, href, last_join = nil)
         if s.match(/\A(#{XREF_ANY}) and (#{XREF_ANY})\z/)
           handle_bares($1, {}, nil, href, " and ")
@@ -85,13 +138,14 @@ module Kramdown
         end
 
         href = href.split(' ')[0] # Remove any trailing (...)
+        target1, target2 = href.split("@", 2)
         multi = last_join != nil
         (sn, s) = s.split(' ', 2)
         loop do
           m = s.match(/\A#{XREF_RE_M}(, (?:and )?| and )?/)
           break if not m
 
-          if not multi and not m[2] and not m[3]
+          if not multi and not m[2] and not m[3] and not target2
             # Modify |attr| if there is a single reference.  This can only be
             # used if there is only one section reference and the section part
             # has no title.
@@ -109,9 +163,13 @@ module Kramdown
           multi = true
           s[m[0]] = ''
 
-          attr1 = { 'target' => href, 'section' => m[1], 'sectionFormat' => 'bare', 'text' => m[2] }
+          attr1 = { 'target' => target1, 'section' => m[1], 'sectionFormat' => 'bare', 'text' => m[2] }
           @tree.children << Element.new(:xref, nil, attr1)
-          @tree.children << Element.new(:text, m[3] || last_join || " of ", {})
+          andof = m[3] || last_join || " of "
+          if andof == " of " && target2
+            andof += rfc_mention(target1)
+          end
+          @tree.children << Element.new(:text, andof, {})
         end
       end
 
@@ -134,18 +192,22 @@ module Kramdown
         else
           href = @src[3]
           attr = {}
+          handled_subref = false
           if $options.v3
             # match Section ... of ...; set section, sectionFormat
             case href.gsub(/[\u00A0\s]+/, ' ') # may need nbsp and/or newlines
             when /\A(#{SECTIONS_RE}) of (.*)\z/
               href = $2
               handle_bares($1, attr, "of", href)
+              handled_subref = true
             when /\A(.*), (#{SECTIONS_RE})\z/
               href = $1
               handle_bares($2, attr, "comma", href)
+              handled_subref = true
             when /\A(.*) \((#{SECTIONS_RE})\)\z/
               href = $1
               handle_bares($2, attr, "parens", href)
+              handled_subref = true
             when /#{XREF_RE_M}<(.+)\z/
               href = $3
               if $2
@@ -167,6 +229,13 @@ module Kramdown
             href = $1
             attr['text'] = $2
           end
+          target1, target2 = href.split("@", 2) # should do this only for sectionref...
+          if target2
+            href = target2
+            unless handled_subref
+              @tree.children << Element.new(:text, rfc_mention(target1), {})
+            end
+          end
           href = self.class.idref_cleanup(href)
           attr['target'] = href
           el = Element.new(:xref, nil, attr)
@@ -260,6 +329,9 @@ module Kramdown
         if anchor = a.delete('href')
           a['target'] = ::Kramdown::Parser::RFC2629Kramdown.idref_cleanup(anchor)
         end
+        if lang = a.delete('lang-')
+          a['xml:lang'] = lang
+        end
         if av = a.delete('noabbrev')      # pseudo attribute -> opts
           opts = opts.merge(noabbrev: TRUTHY[av]) # updated copy
         end
@@ -312,15 +384,28 @@ module Kramdown
       def initialize(*doc)
         super
         @sec_level = 1
+        @location_delta = 100000 # until reset
+        @location_correction = 0 # pre-scanning corrections
         @in_dt = 0
         @footnote_names_in_use = {}
       end
 
-      def convert(el, indent = -INDENTATION, opts = {})
+      def correct_location(location)
+        location + @location_delta + @location_correction
+      end
+
+      def convert(el)
+        opts = el.options[:options]
+        # warn "** tree opts #{opts.inspect}"
+        if nested_ol_types = @options[:nested_ol_types]
+          opts[:nested_ol_types] ||= nested_ol_types
+          # warn "** tree opts out #{opts.inspect}"
+        end
+        indent = -INDENTATION
         if el.children[-1].type == :raw
           raw = convert1(el.children.pop, indent, opts)
         end
-        "#{convert1(el, indent, opts)}#{end_sections(1, indent)}#{raw}"
+        "#{convert1(el, indent, opts)}#{end_sections(1, indent, el.options[:location])}#{raw}"
       end
 
       def convert1(el, indent, opts = {})
@@ -360,10 +445,19 @@ module Kramdown
         generate_id(value).gsub(/-+/, '-')
       end
 
-      def self.process_markdown(v)             # Uuh.  Heavy coupling.
+      def self.process_markdown1(v)             # Uuh.  Heavy coupling.
         doc = ::Kramdown::Document.new(v, $global_markdown_options)
         $stderr.puts doc.warnings.to_yaml unless doc.warnings.empty?
-        doc.to_rfc2629[3..-6] # skip <t>...</t>\n
+        doc.to_rfc2629
+      end
+
+      def self.process_markdown(v)
+        process_markdown1(v)[3..-6] # skip <t>...</t>\n
+      end
+
+      def self.process_markdown_to_rexml(v)
+        s = process_markdown1(v)
+        REXML::Document.new(s)
       end
 
       SVG_COLORS = Hash.new {|h, k| k}
@@ -538,11 +632,10 @@ COLORS
                                             stdin_data: result);
           err << err1.to_s
         when "math", "math-asciitex"
-          math = Shellwords.escape(result)
           result1, err, _s = Open3.capture3("tex2svg --font STIX --speech=false#{svg_opt} #{Shellwords.escape(' ' << result)}");
           begin
             raise Errno::ENOENT if t == "math-asciitex"
-            result, err1, s = Open3.capture3("utftex #{math}#{txt_opt}")
+            result, err1, s = Open3.capture3("utftex -m #{txt_opt}", stdin_data: result)
             if s.exitstatus != 0
               warn "** utftex: #{err1.inspect}"
               raise Errno::ENOENT
@@ -560,7 +653,11 @@ COLORS
         if result1
           result1 = svg_clean(result1) unless dont_clean
           unless dont_check
-            result1, err, _s = Open3.capture3("svgcheck -Xqa", stdin_data: result1);
+            file = Tempfile.new("kramdown-rfc")
+            file.write(result1)
+            file.close
+            result1, err, _s = Open3.capture3("svgcheck -qa #{file.path}");
+            file.unlink
             # warn ["svgcheck:", result1.inspect]
             capture_croak("svgcheck", err)
           end
@@ -651,36 +748,86 @@ COLORS
               if anchor = el.attr['anchor']
                 "##{anchor}"
               elsif lineno = el.options[:location]
-                "approx. line #{lineno}" # XXX
+                "#{correct_location(lineno)}"
               else
                 "UNKNOWN"
               end
+            preprocs = el.attr.delete("pre")
+            checks = el.attr.delete("check")
+            postprocs = el.attr.delete("post")
             case t
+            when "cbor"
+              warn "** There is no sourcecode-type “cbor”."
+              warn "**   Do you mean “cbor-diag” (diagnostic notation)"
+              warn "**   or “cbor-pretty” (annotated hex-dump)?"
             when "json"
-              begin
-                JSON.load(result)
-              rescue => e
-                err1 = "*** #{loc_str}: JSON isn't: #{e.message[0..40]}\n"
-                begin
-                  JSON.load("{" << result << "}")
-                rescue => e
-                  warn err1 << "***  not even with braces added around: #{e.message[0..40]}"
-                end
-              end
-            when "json-from-yaml"
-              begin
-                y = YAML.safe_load(result, aliases: true, filename: loc_str)
-                result = JSON.pretty_generate(y)
-                t = "json"      # XXX, this could be another format!
-              rescue => e
-                warn "*** YAML isn't: #{e.message}\n"
-              end
+              checks ||= "json"
+            when /\A(.*)-from-yaml\z/
+              t = $1
+              preprocs ||= "yaml2json"
             end
+            preprocs = (preprocs || '').split("-")
+            checks = (checks || '').split("-")
+            postprocs = (postprocs || '').split("-")
+            result = sourcecode_checkproc(preprocs, checks, postprocs, loc_str, result)
             "#{' '*indent}<figure#{el_html_attributes(el)}><#{gi}#{html_attributes(artwork_attr)}><![CDATA[#{result}#{result =~ /\n\Z/ ? '' : "\n"}]]></#{gi}></figure>\n"
           end
         end
       end
 
+      def sourcecode_proc(proc, loc_str, result)
+        case proc
+        when "dedent"
+          result = remove_indentation(result)
+        when /\Afold(\d*)(left(\d*))?(dry)?\z/
+          fold = [$1.to_i,            # col 0 for ''
+                  ($3.to_i if $2),    # left 0 for '', nil if no "left"
+                  $4]                 # dry
+          result = fix_unterminated_line(fold8792_1(trim_empty_lines_around(result), *fold)) # XXX
+        when "yaml2json"
+          begin
+            y = YAML.safe_load(result, aliases: true, filename: loc_str)
+            result = JSON.pretty_generate(y)
+          rescue => e
+            warn "*** #{loc_str}: YAML isn't: #{e.message}\n"
+          end
+        else
+          warn "*** #{loc_str}: unknown proc '#{proc}'"
+        end
+        result
+      end
+
+      def sourcecode_checkproc(preprocs, checks, postprocs, loc_str, result)
+        preprocs.each do |proc|
+          result = sourcecode_proc(proc, loc_str, result)
+        end if preprocs
+        check_input = result
+        checks.each do |check|
+          case check
+          when "skipheader"
+            check_input = handle_artwork_sourcecode(check_input).sub(/.*?\n\n/m, '')
+          when "json"
+            # check for 8792; undo if needed:
+            begin
+              JSON.load(handle_artwork_sourcecode(check_input))
+            rescue => e
+              err1 = "*** #{loc_str}: JSON isn't: #{JSON.dump(e.message[0..40])}\n"
+              begin
+                JSON.load("{" << check_input << "}")
+              rescue => e
+                warn err1 << "***  not even with braces added around: #{JSON.dump(e.message[0..40])}"
+              end
+            end
+          else
+            warn "*** #{loc_str}: unknown check '#{check}'"
+          end
+        end if checks
+        postprocs.each do |proc|
+          result = sourcecode_proc(proc, loc_str, result)
+        end if postprocs
+        result
+      end
+
       def mk_artwork(artwork_attr, typ, content)
         "<artwork #{html_attributes(artwork_attr.merge("type" => typ))}>#{content}</artwork>"
       end
@@ -700,7 +847,7 @@ COLORS
         end
       end
 
-      def end_sections(to_level, indent)
+      def end_sections(to_level, indent, location)
         if indent < 0
           indent = 0
         end
@@ -709,7 +856,7 @@ COLORS
           @sec_level = to_level
           "#{' '*indent}</section>\n" * delta
         else
-          $stderr.puts "Incorrect section nesting: Need to start with 1"
+          $stderr.puts "** #{correct_location(location)}: Bad section nesting: start heading level at 1 and increment by 1"
         end
       end
 
@@ -761,7 +908,7 @@ COLORS
         clean, irefs = clean_pcdata(inner_a(el, indent, opts))
         el.attr['title'] = clean
         end
-        "#{end_sections(el.options[:level], indent)}#{' '*indent}<section#{@sec_level += 1; el_html_attributes(el)}>#{irefs}\n"
+        "#{end_sections(el.options[:level], indent, el.options[:location])}#{' '*indent}<section#{@sec_level += 1; el_html_attributes(el)}>#{irefs}\n"
       end
 
       def convert_hr(el, indent, opts) # misuse for page break
@@ -779,7 +926,26 @@ COLORS
           "#{' '*indent}<t><list#{attrstring}>\n#{inner(el, indent, opts)}#{' '*indent}</list></t>\n"
         end
       end
-      alias :convert_ol :convert_ul
+
+      def convert_ol(el, indent, opts)
+        if @options[:ol_start_at_first_marker] and (first_list_marker =
+                                                    el.options[:first_list_marker])
+          el.attr['start'] ||= first_list_marker[/\d+/]
+        end
+        nested_types = opts[:nested_ol_types] || ["1"]
+        # warn "** ol opts #{opts.inspect} types #{nested_types.inspect}"
+        if nested_attr = el.attr.delete('nestedOlTypes')
+          nested_types = ::Kramdown::Options.parse(:nested_ol_types, nested_attr)
+        end
+        if nested_types = nested_types.dup
+          # warn "** nested_types #{nested_types.inspect}"
+          nested_here = nested_types.shift
+          opts = opts.merge(nested_ol_types: nested_types << nested_here)
+          el.attr['type'] ||= nested_here
+          # warn "** actual ol type #{el.attr['type'].inspect}"
+        end
+        convert_ul(el, indent, opts)
+      end
 
       def convert_dl(el, indent, opts)
         if $options.v3
@@ -878,6 +1044,15 @@ COLORS
       end
 
       def convert_xml_comment(el, indent, opts)
+        if el.value =~ /\A<\?line (([-+]?)[0-9]+)\?>\z/
+          lineno = $1.to_i
+          case $2
+          when ''               # absolute
+            @location_delta = lineno - el.options[:location]
+          when '+', '-'         # correction (pre-scanning!)
+            @location_correction += lineno
+          end
+        end
         if el.options[:category] == :block && !el.options[:parent_is_raw]
           ' '*indent + el.value + "\n"
         else
@@ -1114,7 +1289,8 @@ COLORS
       XML_RESOURCE_ORG_MAP = {
         "RFC" => ["bibxml", 86400*7, false,
                   ->(fn, n){ [name = "reference.RFC.#{"%04d" % n.to_i}.xml",
-                              "https://www.rfc-editor.org/refs/bibxml/#{name}"] }
+                              "https://bib.ietf.org/public/rfc/bibxml/#{name}"] }
+# was                         "https://www.rfc-editor.org/refs/bibxml/#{name}"] }
                  ],
         "I-D" => ["bibxml3", false, false,
                   ->(fn, n){ [fn,
@@ -1165,6 +1341,7 @@ COLORS
           end
         end
         if alt == ":include:"   # Really bad misuse of tag...
+          ann = el.attr.delete('ann')
           anchor = el.attr.delete('anchor') || (
             # not yet
             warn "*** missing anchor for '#{src}'"
@@ -1196,6 +1373,7 @@ COLORS
               d = REXML::Document.new(to_insert)
               d.xml_decl.nowrite
               d.delete d.doctype
+              d.context[:attribute_quote] = :quote  # Set double-quote as the attribute value delimiter
               d.root.attributes["anchor"] = anchor
               if t == "RFC" or t == "I-D"
                 if KRAMDOWN_NO_TARGETS || !KRAMDOWN_KEEP_TARGETS
@@ -1212,6 +1390,11 @@ COLORS
               elsif t == "IANA"
                 d.root.attributes["target"].sub!(%r{\Ahttp://www.iana.org/assignments/}, 'https://www.iana.org/assignments/')
               end
+              if ann
+                el = ::Kramdown::Converter::Rfc2629::process_markdown_to_rexml(ann).root
+                el.name = "annotation"
+                d.root.add_element(el)
+              end
               to_insert = d.to_s
             rescue Exception => e
               warn "** Can't manipulate reference XML: #{e}"
@@ -1270,7 +1453,7 @@ COLORS
       end
 
       def convert_raw(el, indent, opts)
-        end_sections(1, indent) +
+        end_sections(1, indent, el.options[:location]) +
         el.value + (el.options[:category] == :block ? "\n" : '')
       end
 
@@ -1376,6 +1559,19 @@ COLORS
         "<iref#{html_attributes(attr)}/>"
       end
 
+      def nobr_hack(s)          # replace this by actual <nobr> once that exists
+        # https://github.com/ietf-tools/xml2rfc/blob/main/xml2rfc/utils.py#L42
+        s.gsub(/([-\s\/])(?!\s)/) { case rep = $1
+                                   when /\A\s\z/
+                                     "\u00A0" # nbsp
+                                   when "-"
+                                     "\u2011" # nbhy -- XXX this might mangle dashes
+                                   else
+                                     "#{rep}\u2060"
+                                   end
+        }
+      end
+
       def convert_iref(el, indent, opts)
         iref_attr(el.attr['target'])
       end
@@ -1398,6 +1594,18 @@ COLORS
           return "<bcp14>#{value}</bcp14>"
         end
 
+        hacked_value = value
+
+        nobr = false
+        if title && title =~ /\A<nobr>(\z|\s)/
+          nobr = true
+          _nobr, title = title.split(' ', 2)
+          hacked_value = nobr_hack(value)
+          if title.nil? || title.empty?
+            return hacked_value # we have "exhausted" this abbrev -- suppress normal meaning
+          end
+        end
+
         if title && title[0] == "#"
           target, title = title.split(' ', 2)
           if target == "#"
@@ -1410,20 +1618,28 @@ COLORS
         end
 
         if item = title
-          m = title.scan(Parser::RFC2629Kramdown::IREF_START)
-          if m.empty?
+          pairs = title.split(Parser::RFC2629Kramdown::IREF_START).each_slice(2).to_a
+          replacement = pairs.map {|x,| s = x.strip; s unless s.empty?}.compact.join(" ")
+          irefs = pairs.map {|_,x| x && [x]}.compact
+          warn "@@@ ABBREV MISMATCH #{irefs}" if title.scan(Parser::RFC2629Kramdown::IREF_START) != irefs
+          if irefs.empty?
             subitem = value
           else
-            iref = m.map{|a,| iref_attr(a)}.join('')
+            iref = irefs.map{|a,| iref_attr(a)}.join('')
+          end
+          unless replacement.empty?
+            replacement = nobr_hack(replacement) if nobr # XXX this can break XML
+            replacement = ::Kramdown::Converter::Rfc2629::process_markdown(replacement)
+            hacked_value = replacement
           end
         else
           item = value
         end
         iref ||= "<iref#{html_attributes(item: item, subitem: subitem)}/>"
         if target
-          "#{iref}<xref#{html_attributes(target: target, format: "none")}>#{value}</xref>"
+          "#{iref}<xref#{html_attributes(target: target, format: "none")}>#{hacked_value}</xref>"
         else
-          "#{iref}#{value}"
+          "#{iref}#{hacked_value}"
         end
       end
 
